@@ -69,6 +69,8 @@ class Interpreter:
         callable file-like object where standard error will be sent.
     use_numpy : bool
         whether to use functions from numpy.
+    max_statement_length : int
+        maximum length of expression allowed [50,000 characters]
     minimal : bool
         create a minimal interpreter: disable all options (see Note 1).
     no_if : bool
@@ -107,8 +109,9 @@ class Interpreter:
     """
 
     def __init__(self, symtable=None, usersyms=None, writer=None,
-                 err_writer=None, use_numpy=True, minimal=False,
-                 no_if=False, no_for=False, no_while=False, no_try=False,
+                 err_writer=None, use_numpy=True,
+                 max_statement_length=50000, minimal=False, no_if=False,
+                 no_for=False, no_while=False, no_try=False,
                  no_functiondef=False, no_ifexp=False, no_listcomp=False,
                  no_augassign=False, no_assert=False, no_delete=False,
                  no_raise=False, no_print=False, max_time=None,
@@ -116,6 +119,7 @@ class Interpreter:
 
         self.writer = writer or stdout
         self.err_writer = err_writer or stderr
+        self.max_statement_length = max(1, min(1.e8, max_statement_length))
 
         if symtable is None:
             if usersyms is None:
@@ -128,6 +132,7 @@ class Interpreter:
         self.error_msg = None
         self.expr = None
         self.retval = None
+        self._calldepth = 0
         self.lineno = 0
         self.start_time = time.time()
         self.use_numpy = HAS_NUMPY and use_numpy
@@ -249,6 +254,10 @@ class Interpreter:
     #  eval:   string statement -> result = run(parse(statement))
     def parse(self, text):
         """Parse statement/expression to Ast representation."""
+        if len(text) > self.max_statement_length:
+            msg = f'length of text exceeds {self.max_statement_length:d} characters'
+            self.raise_exception(None, msg='Runtime Error', expr=msg)
+
         self.expr = text
         try:
             out = ast.parse(text)
@@ -304,38 +313,48 @@ class Interpreter:
         """Call class instance as function."""
         return self.eval(expr, **kw)
 
-    def eval(self, expr, lineno=0, show_errors=True):
+    def eval(self, expr, lineno=0, show_errors=True, raise_errors=False):
         """Evaluate a single statement."""
         self.lineno = lineno
         self.error = []
         self.start_time = time.time()
-        try:
-            node = self.parse(expr)
-        except:
-            errmsg = exc_info()[1]
-            if len(self.error) > 0:
-                errmsg = "\n".join(self.error[0].get_error())
-            if not show_errors:
-                try:
-                    exc = self.error[0].exc
-                except:
-                    exc = RuntimeError
-                raise exc(errmsg)
-            print(errmsg, file=self.err_writer)
-            return
+        if isinstance(expr, str):
+            if len(expr) > self.max_statement_length:
+                msg = f'length of text exceeds {self.max_statement_length:d} characters'
+                raise ValueError(msg)
+
+            try:
+                node = self.parse(expr)
+            except:
+                errmsg = exc_info()[1]
+                if len(self.error) > 0:
+                    errmsg = "\n".join(self.error[0].get_error())
+                if raise_errors:
+                    try:
+                        exc = self.error[0].exc
+                    except:
+                        exc = RuntimeError
+                    raise exc(errmsg)
+                if show_errors:
+                    print(errmsg, file=self.err_writer)
+                return
+        else:
+            node = expr
+
         try:
             return self.run(node, expr=expr, lineno=lineno)
         except:
             errmsg = exc_info()[1]
             if len(self.error) > 0:
                 errmsg = "\n".join(self.error[0].get_error())
-            if not show_errors:
+            if raise_errors:
                 try:
                     exc = self.error[0].exc
                 except:
                     exc = RuntimeError
                 raise exc(errmsg)
-            print(errmsg, file=self.err_writer)
+            if show_errors:
+                print(errmsg, file=self.err_writer)
             return
 
     @staticmethod
@@ -354,6 +373,8 @@ class Interpreter:
 
     def on_return(self, node):  # ('value',)
         """Return statement: look for None, return special sentinel."""
+        if self._calldepth == 0:
+            raise SyntaxError('cannot return at top level')
         self.retval = self.run(node.value)
         if self.retval is None:
             self.retval = ReturnedNone
@@ -744,13 +765,20 @@ class Interpreter:
         if kwargs is not None:
             keywords.update(self.run(kwargs))
 
+        if isinstance(func, Procedure):
+            self._calldepth += 1
         try:
-            return func(*args, **keywords)
+            out = func(*args, **keywords)
         except Exception as ex:
+            out = None
             func_name = getattr(func, '__name__', str(func))
             self.raise_exception(
                 node, msg="Error running function call '%s' with args %s and "
                 "kwargs %s: %s" % (func_name, args, keywords, ex))
+        finally:
+            if isinstance(func, Procedure):
+                self._calldepth -= 1
+        return out
 
     def on_arg(self, node):    # ('test', 'msg')
         """Arg for function definitions."""
@@ -927,6 +955,7 @@ class Procedure:
         save_symtable = self.__asteval__.symtable.copy()
         self.__asteval__.symtable.update(symlocals)
         self.__asteval__.retval = None
+        self.__asteval__._calldepth += 1
         retval = None
 
         # evaluate script of function
@@ -942,5 +971,6 @@ class Procedure:
                 break
 
         self.__asteval__.symtable = save_symtable
+        self.__asteval__._calldepth -= 1
         symlocals = None
         return retval
